# YouTube
YT = "https://www.youtube.com/"
YT_VID_URL = YT + "watch?v="
YT_CHANNEL_URL = YT + "channel/"

# Telega.ph
TELEGRA_PH = "https://telegra.ph"
TELEGRA_PH_HEADERS = {
    "authority": "telegra.ph",
    "accept": "application/json, text/javascript, */*; q=0.01",
    "x-requested-with": "XMLHttpRequest",
    "user-agent": "Python_telegraph_poster/0.1",
    "origin": TELEGRA_PH,
    "referer": f"{TELEGRA_PH}/",
    "accept-language": "en-US,en;q=0.9",
}



__all__ = ["Downloader"]


import asyncio
import logging
import os
import time

from math import floor
from typing import Any, Callable, Dict, Union

import yt_dlp as youtube_dl

from pyrogram import ContinuePropagation, StopPropagation, StopTransmission
from pyrogram.errors import FloodWait, MessageNotModified
from pyrogram.types import CallbackQuery, Message

# from youtube_dl.utils import DownloadError, GeoRestrictedError
from yt_dlp.utils import DownloadError, GeoRestrictedError

from iytdl.processes import Process
from iytdl.utils import *


logger = logging.getLogger(__name__)


class Downloader:
    async def video_downloader(
        self, url: str, uid: str, rnd_key: str, prog_func: Callable
    ) -> Union[int, str]:

        options = {
            "addmetadata": True,
            "geo_bypass": True,
            "nocheckcertificate": True,
            "outtmpl": os.path.join(
                str(self.download_path), rnd_key, "%(title)s-%(format)s.%(ext)s"
            ),
            "logger": logger,
            "progress_hooks": [prog_func],
            "format": uid,
            "writethumbnail": True,
            "prefer_ffmpeg": True,
            "postprocessors": [{"key": "FFmpegMetadata"}],
            "quiet": self.silent,
            "logtostderr": self.silent,
        }

        if (ext_dl := self.external_downloader) is not None:
            options.update(ext_dl._export())
        return await self.ytdownloader(url, options)

    async def audio_downloader(
        self, url: str, uid: str, rnd_key: str, prog_func: Callable
    ) -> Union[int, str]:
        options = {
            "outtmpl": os.path.join(
                str(self.download_path), rnd_key, "%(title)s-%(format)s.%(ext)s"
            ),
            "logger": logger,
            "progress_hooks": [prog_func],
            "writethumbnail": True,
            "prefer_ffmpeg": True,
            "format": "bestaudio/best",
            "geo_bypass": True,
            "nocheckcertificate": True,
            "postprocessors": [
                {
                    "key": "FFmpegExtractAudio",
                    "preferredcodec": "mp3",
                    "preferredquality": uid,
                },
                {"key": "EmbedThumbnail"},
                {"key": "FFmpegMetadata"},
            ],
            "quiet": self.silent,
            "logtostderr": self.silent,
        }
        if (ext_dl := self.external_downloader) is not None:
            options.update(ext_dl._export())
        return await self.ytdownloader(url, options)

    @run_sync
    def ytdownloader(self, url: str, options: Dict[str, Any]) -> Union[int, str]:
        try:
            with youtube_dl.YoutubeDL(options) as ytdl:
                return ytdl.download([url])
        except DownloadError:
            logger.error("[DownloadError] : Failed to Download Video")
        except GeoRestrictedError:
            logger.error(
                "[GeoRestrictedError] : The uploader has not made this video"
                " available in your country"
            )
        except Exception:
            logger.exception("Something Went Wrong")

    async def download(
        self,
        url: str,
        uid: str,
        downtype: str,
        update: Union[Message, CallbackQuery],
        with_progress: bool = True,
        edit_rate: int = 8,
    ) -> str:
        """Download Media with progress bar

        Parameters:
        ----------
            - url (`str`): Youtube_dl supported URL.
            - uid (`str`): Preferred media choice.
            - downtype (`str`): [`'audio'` | `'video'`].
            - update (`Union[Message, CallbackQuery]`): A Pyrogram update to display progress.
            - with_progress (`bool`, optional): Enable / Disable progress. (Defaults to `True`)
            - edit_rate (`int`, optional): Progress edit rate in seconds. (Defaults to `8`)

        Returns:
        -------
            `str`: Key to upload media, After successful download

        Raises:
        ------
            `TypeError`: On unsupported `downtype`
            `StopTransmission`: When download is cancelled
            `DownloadError`: In case youtube_dl download return code is not equal to 0
        """
        last_update_time = None

        process = Process(update)
        key = rnd_key()

        def prog_func(prog_data: Dict) -> None:
            nonlocal last_update_time
            now = int(time.time())
            # Only edit message once every 8 seconds to avoid ratelimits

            if process.is_cancelled:
                logger.warning("Download process is Cancelled")
                raise StopTransmission

            if prog_data.get("status") == "finished":
                progress = "🔄  Download finished, Uploading..."
            elif last_update_time is None or (now - last_update_time) >= edit_rate:
                # ------------ Progress Data ------------ #
                if not (
                    (eta := prog_data.get("eta")) and (speed := prog_data.get("speed"))
                ):
                    return
                current = prog_data.get("downloaded_bytes")
                filename = prog_data.get("filename")
                if total := prog_data.get("total_bytes"):
                    percentage = round(current / total * 100)
                    progress_bar = (
                        f"[{'█' * floor(15 * percentage / 100)}"
                        f"{'░' * floor(15 * (1 - percentage / 100))}]"
                    )
                    # ---------------------------------------- #
                    progress = f"""
<i>Downloading:</i>  <code>{filename}</code>
<b>Completed:</b>  <code>{humanbytes(current)} / {humanbytes(total)}</code>
<b>Progress:</b>  <code>{progress_bar} {percentage} %</code>
<b>Speed:</b>  <code>{humanbytes(speed)}</code>
<b>ETA:</b>  <code>{time_formater(eta)}</code>
"""
                else:
                    # Total is None, Generic progress bar
                    progress = f"""
<i>Downloading:</i>  <code>{filename}</code>
<b>Completed:</b>  <code>{humanbytes(current)} / [N/A]</code>
<b>Speed:</b>  <code>-</code>
<b>ETA:</b>  <code>-</code>
"""
            else:
                return
            if with_progress:
                self.loop.create_task(self.progress_func(process, progress))
            last_update_time = now

        if downtype == "video":
            out = await self.video_downloader(url, uid, key, prog_func)
        elif downtype == "audio":
            out = await self.audio_downloader(url, uid, key, prog_func)
        else:
            raise TypeError(f"'{downtype}' is Unsupported !")

        if isinstance(out, int) and out == 0:
            return key
        raise DownloadError(str(out))

    @staticmethod
    async def progress_func(process: Process, text: str) -> None:
        try:
            await process.edit(
                text=text,
                parse_mode="HTML",
                disable_web_page_preview=True,
                reply_markup=process.cancel_markup,
            )
        except FloodWait as f:
            await asyncio.sleep(f.x + 2)
        except (StopPropagation, StopTransmission) as p_e:
            raise p_e
        except (ContinuePropagation, MessageNotModified):
            pass
        except Exception:
            logger.exception("Progress")



__all__ = ["iYTDL"]

import asyncio
import hashlib
import re

from pathlib import Path
from typing import Optional, Tuple, Union

from aiohttp import ClientSession
from html_telegraph_poster import TelegraphPoster
from pyrogram.types import InlineKeyboardButton, InlineKeyboardMarkup, InputMediaPhoto
from youtubesearchpython.__future__ import VideosSearch

from iytdl import types
from iytdl.constants import YT_VID_URL
from iytdl.downloader import Downloader
from iytdl.exceptions import *  # noqa ignore=F405
from iytdl.extractors import Extractor
from iytdl.formatter import ResultFormatter, gen_search_markup
from iytdl.sql_cache import AioSQLiteDB
from iytdl.upload_lib.uploader import Uploader
from iytdl.utils import run_sync


class iYTDL(Extractor, Downloader, Uploader):
    def __init__(
        self,
        log_group_id: Union[int, str],
        session: Optional[ClientSession] = None,
        silent: bool = False,
        download_path: str = "downloads",
        loop: Optional[asyncio.AbstractEventLoop] = None,
        default_thumb: str = "https://i.imgur.com/4LwPLai.png",
        cache_path: str = "",
        delete_media: bool = False,
        external_downloader: Optional[types.ExternalDownloader] = None,
    ) -> None:
        """Main Class

        Parameters:
        ----------
            - log_group_id (`Union[int, str]`):  Log channel / group id to upload downloaded media.
            - session (`Optional[ClientSession]`, optional): Aiohttp ClientSession. (Defaults to `None`)
            - silent (`bool`, optional): Disable youtube_dl stdout. (Defaults to `False`)
            - download_path (`str`, optional): Custom download location. (Defaults to `"downloads"`)
            - loop (`Optional[asyncio.AbstractEventLoop]`, optional): Event loop. (Defaults to `None`)
            - default_thumb (`str`, optional): Fallback thumbnail. (Defaults to `"https://i.imgur.com/4LwPLai.png"`)
            - cache_path (`str`, optional): Path to store cache. (Defaults to `""`)
            - delete_media: (`bool`, optional): Delete media from local storage after uploading on Telegram. (Defaults to `False`)
            - external_downloader: (`Optional[types.ExternalDownloader]`, optional): External Downloader e.g `types.external_downloader.Aria2c`. (Defaults to `None`)
        """
        self.yt_link_regex = re.compile(
            r"(?:youtube\.com|youtu\.be)/(?:[\w-]+\?v=|embed/|v/|shorts/)?([\w-]{11})"
        )
        self.generic_url_regex = re.compile(r"^https?://\S+")
        self.default_thumb = default_thumb
        self.http = session or ClientSession()
        _cache_path = Path(cache_path)
        if _cache_path.is_file():
            raise TypeError(f"'{cache_path}' expected a Directory got a File instead")
        self.cache = AioSQLiteDB(
            _cache_path.joinpath("yt_search_cache.db"), clean=False
        )
        self.loop = loop or asyncio.get_event_loop()
        self.download_path = Path(download_path)
        self.log_group_id = log_group_id

        self.download_path.mkdir(exist_ok=True, parents=True)
        self.external_downloader = external_downloader
        self.delete_file_after_upload = delete_media

        super().__init__(silent=silent)

    @classmethod
    async def init(cls, *args, **kwargs) -> "iYTDL":
        yt = cls(*args, **kwargs)
        await yt.start()
        return yt

    async def search(self, query: str) -> types.SearhResult:
        """Search

        Parameters:
        ----------
            - query (`str`): Provide URL or text.

        Raises:
        ------
            `NoResultFoundError`: In case of no result.

        Returns:
        -------
            `types.SearhResult`
        """
        hash_key = hashlib.sha1(query.encode(encoding="UTF-8")).hexdigest()
        # sqlite doesn't support numbers in table name
        key = re.sub(r"\d+", "", hash_key)[:10]
        if cached_data := await self.cache.get_key(key, index=0):
            s_len, v_data = cached_data
        else:
            videosResult = await VideosSearch(query, limit=15).next()
            if len((res := videosResult["result"])) == 0:
                raise NoResultFoundError
            search_data = await asyncio.gather(
                *map(lambda x: ResultFormatter.parse(self, x), res[:15])
            )
            await self.cache.set_key(key, search_data)
            v_data = search_data[0]
            s_len = len(search_data)
        r1 = ResultFormatter(**v_data)
        return types.SearhResult(
            key, r1.msg, r1.thumb, gen_search_markup(key, r1.yt_id, s_len)
        )

    async def next_result(self, key: str, index: int) -> types.SearhResult:
        """Get next result from cached data

        Parameters:
        ----------
            - key (`str`): Unique Key.
            - index (`int`): Result Index.

        Returns:
        -------
            `types.SearhResult`
        """
        if cached_data := await self.cache.get_key(key, index=index - 1):
            s_len, v_data = cached_data
            vid = ResultFormatter(**v_data)
            return types.SearhResult(
                key,
                vid.msg,
                vid.thumb,
                gen_search_markup(key, vid.yt_id, s_len, index),
            )

    async def extract_info_from_key(self, key: str) -> Optional[types.SearhResult]:
        """
        Parameters:
        ----------
            - key (`str`): Unique Key.

        Returns:
        -------
            `Optional[types.SearhResult]`: If key exist in cache.
        """
        if len(key) == 11:
            # yt_id
            return await self.get_download_button(key)
        if url := await self.cache.get_url(key):
            return await self.generic_extractor(key, url)

    async def parse(self, search_query: str, extract: bool = True) -> types.SearhResult:
        """Automatically parses `search_query`.

        Parameters:
        ----------
            - search_query (`str`): accepts [Youtube URL | URL | Text].
            - extract (`bool`, optional): Extract Information (Defaults to `True`)

        Raises:
        ------
            `NoResultFoundError`: In case of no result.

        Returns:
        -------
            `types.SearhResult`
        """
        query_split = search_query.split()
        if len(query_split) == 1:
            url = query_split[0]
            # can be some text or an URL
            if match := self.yt_link_regex.search(url):
                # youtube link
                yt_id = match.group(1)
                thumb = await self.get_ytthumb(yt_id)
                if extract:
                    return await self.get_download_button(yt_id)
                return types.SearhResult(
                    yt_id,
                    f"**[YouTube URL]** -> `'{YT_VID_URL}{yt_id}'`",
                    thumb,
                    InlineKeyboardMarkup(
                        [
                            [
                                InlineKeyboardButton(
                                    "⚙️ Extract Info",
                                    callback_data=f"yt_extract_info|{yt_id}",
                                )
                            ]
                        ]
                    ),
                )
            elif self.generic_url_regex.search(url):
                # Matches URL regex that can be supported by YoutubeDL
                key = await self.cache.save_url(url)
                if extract:
                    return await self.generic_extractor(key, url)

                return types.SearhResult(
                    key,
                    f"**[Generic URL]** -> `'{url}'`",
                    self.default_thumb,
                    InlineKeyboardMarkup(
                        [
                            [
                                InlineKeyboardButton(
                                    "⚙️ Extract Info",
                                    callback_data=f"yt_extract_info|{key}",
                                )
                            ]
                        ]
                    ),
                )
        # YT Search if query didn't matched earlier or is of multiple words
        return await self.search(search_query.strip())

    async def listview(self, key: str) -> Tuple[InputMediaPhoto, InlineKeyboardMarkup]:
        """List data on Telegra.ph

        Parameters:
        ----------
            - key (`str`): Unique Key.

        Returns:
        -------
            `Tuple[InputMediaPhoto, InlineKeyboardMarkup]`
        """
        if cached_data := await self.cache.get_key(key):
            content = "\n".join(
                map(
                    lambda x: ResultFormatter(*x[1]).list_view(x[0]),
                    enumerate(cached_data, start=1),
                )
            )
            telegraph = await self.paste_to_tg("📜  LIST VIEW", content)
            buttons = InlineKeyboardMarkup(
                [
                    [
                        InlineKeyboardButton(
                            "↗️  Click To Open",
                            url=telegraph,
                        )
                    ],
                    [
                        InlineKeyboardButton(
                            "📰  Detailed View",
                            callback_data=f"yt_next|{key}|0",
                        )
                    ],
                ]
            )
            image = ResultFormatter(*cached_data[0]).thumb
            return InputMediaPhoto(media=image), buttons

    @staticmethod
    async def paste_to_tg(title: str, content: str) -> str:
        """Paste to Telegra.ph

        Parameters:
        ----------
            - title (`str`): Page title.
            - content (`str`): Page content.

        Returns:
        -------
            `str`: Telegra.ph URL
        """
        post_client = TelegraphPoster(use_api=True)
        auth_name = "X"
        post_client.create_api_token(auth_name)
        return (
            await run_sync(post_client.post)(
                title=title,
                author=auth_name,
                author_url="https://t.me/x_xtests",
                text=content,
            )
        ).get("url")

    async def get_ytthumb(self, yt_id: str) -> str:
        """Get YouTube video thumbnail from video ID

        Parameters:
        ----------
            - yt_id (`str`): YouTube video ID.

        Returns:
        -------
            `str`: Thumbnail URL
        """
        for quality in (
            "maxresdefault",
            "hqdefault",
            "sddefault",
            "mqdefault",
            "default",
        ):
            link = f"https://i.ytimg.com/vi/{yt_id}/{quality}.jpg"
            async with self.http.get(link) as resp:
                if resp.status == 200:
                    break
        else:
            link = self.default_thumb
        return link

    async def stop(self) -> None:
        """Stop iYTDL instance manually or Use Context Manager"""
        if self.http and not self.http.closed:
            await self.http.close()
        await self.cache.close()

    async def start(self) -> None:
        """Start iYTDL instance manually or Use Context Manager"""
        await self.cache._init()

    async def __aenter__(self) -> "iYTDL":
        await self.start()
        return self

    async def __aexit__(self, *_, **__) -> None:
        await self.stop()



__all__ = ["audio", "video", "photo"]

from typing import Tuple


audio: Tuple[str] = (".mp3", ".flac", ".wav", ".m4a")
video: Tuple[str] = (".mkv", ".mp4", ".webm")
photo: Tuple[str] = (".jpg", ".jpeg", ".png", ".bmp", ".webp")



__all__ = ["AioSQLiteDB"]

import os

from typing import Any, Dict, List, Optional, Tuple, Union

import aiosqlite

from iytdl.utils import rnd_key


class AioSQLiteDB:
    db: aiosqlite.Connection
    cur: aiosqlite.Cursor

    def __init__(self, db_name: str, clean: bool = False) -> None:
        """Create / Load Cache

        Parameters:
        ----------
            db_name (`str`): Cache file name.

            clean (`bool`, optional): Delete old cache and create new. (Defaults to `False`)

        """
        if clean and os.path.isfile(db_name):
            os.remove(db_name)
        self.db_name = db_name

    async def _init(self) -> None:
        """Async init"""
        try:
            self.con = await aiosqlite.connect(self.db_name)
        except aiosqlite.OperationalError:
            # DB is corrupt
            if os.path.isfile(self.db_name):
                os.remove(self.db_name)
            self.con = await aiosqlite.connect(self.db_name)
        self.cur = await self.con.cursor()
        await self.__init_tables()

    async def __init_tables(self) -> None:
        """Create required Tables"""
        await self.cur.execute(
            """
CREATE TABLE IF NOT EXISTS url_cache (
    key TEXT NOT NULL UNIQUE,
    url TEXT,
    PRIMARY KEY(key)
);"""
        )
        await self.con.commit()

    async def set_key(self, key: str, value: List[Dict[str, Any]]) -> None:
        """Set Key in Cache

        Parameters:
        ----------
            key (`str`): Unique Key.

            value (`List[Dict[str, Any]]`): YT Search Data.

        """
        await self.cur.execute(
            f"""
CREATE TABLE IF NOT EXISTS {key} (
    yt_id TEXT NOT NULL UNIQUE,
    thumb TEXT,
    title TEXT,
    body TEXT,
    duration TEXT,
    views	TEXT,
    upload_date TEXT,
    chnl_name TEXT,
    chnl_id	TEXT,
    PRIMARY KEY(yt_id)
);"""
        )
        params = list(value[0].keys())
        cmd = (
            f"INSERT INTO "
            f"{key}({', '.join(params)}) "
            f"VALUES({', '.join(['?' for _ in range(len(params))])}) "
        )
        try:
            await self.cur.executemany(
                cmd,
                [tuple(x.values()) for x in value],
            )
        except aiosqlite.IntegrityError:
            pass
        await self.con.commit()

    async def get_key(
        self, key: str, index: Optional[int] = None
    ) -> Union[Tuple[int, str, None], Dict[str, str], None]:
        """Get Data saved in Cache from Key

        Parameters:
        ----------
            key (`str`): Unique Key.

            index (`Optional[int]`, optional): Result index. (Defaults to `None`)

        Returns:
        -------
            Union[Tuple[int, str, None], Dict[str, str], None]
        """
        try:
            await self.cur.execute(f"SELECT * FROM {key}")
        except aiosqlite.OperationalError:
            return
        data = await self.cur.fetchall()
        d_len = len(data)
        if index is None:
            return data
        if 0 <= index < d_len:
            return d_len, dict(
                zip(map(lambda x: x[0], self.cur.description), data[index])
            )

    async def save_url(self, url: str) -> str:
        """Save Url and get Key.

        Parameters:
        ----------
            url (`str`): Http URL.

        Returns:
        -------
            str: Unique Key

        """
        # Check Existing Key
        await self.cur.execute("SELECT key FROM url_cache WHERE url = ?", (url,))
        if old_key := await self.cur.fetchone():
            return old_key[0]
        # New Key
        key = rnd_key(5)
        await self.cur.execute(
            "INSERT INTO url_cache(key, url) VALUES(?, ?)", (key, url)
        )
        await self.con.commit()
        return key

    async def get_url(self, key: str) -> Optional[str]:
        """Get Saved URL from Key

        Parameters:
        ----------
            key (`str`): Unique Key.

        Returns:
        -------
            Optional[str]: URL if found

        """
        await self.cur.execute("SELECT url FROM url_cache WHERE key = ?", (key,))
        if value := await self.cur.fetchone():
            return value[0]

    async def close(self) -> None:
        """Close Cache File"""
        await self.con.close()



__all__ = ["ResultFormatter", "gen_search_markup"]

import json

from html import escape
from typing import Any, Dict, Union

from pyrogram.types import InlineKeyboardButton, InlineKeyboardMarkup

import iytdl

from iytdl.constants import *  # noqa ignore=F405


class ResultFormatter:
    def __init__(self, *args, **kwargs) -> None:
        if kwargs:
            items = kwargs.items()
        elif args:
            items = zip(
                (
                    "yt_id",
                    "thumb",
                    "title",
                    "body",
                    "duration",
                    "views",
                    "upload_date",
                    "chnl_name",
                    "chnl_id",
                ),
                args,
            )
        else:
            raise ValueError("No Input was given !")

        for key, value in items:
            setattr(self, key, value or None)

    @classmethod
    async def parse(
        cls: "ResultFormatter", yt_class: "iytdl.iYTDL", raw_result: Dict[str, Any]
    ) -> Dict[str, str]:
        """Formats YouTube Search Results

        Parameters:
        ----------
            - cls (`ResultFormatter`): new class instance.
            - yt_class (`iytdl.iYTDL`).
            - raw_result (`Dict[str, Any]`): Unformatted results.

        Returns:
        -------
            `Dict[str, str]`: On Success
        """

        yt_id = raw_result.get("id")
        return cls(
            yt_id=yt_id,
            # raw_result.get("thumbnails")[-1].get("url"),
            # Below thumb works everytime unlike telegra.ph or the above link.
            thumb=await yt_class.get_ytthumb(yt_id),
            title=raw_result.get("title"),
            body="".join(map(lambda x: x["text"], descripition))
            if (descripition := raw_result.get("descriptionSnippet"))
            else None,
            duration=acc.get("duration")
            if (acc := raw_result.get("accessibility"))
            else None,
            views=view.get("short") if (view := raw_result.get("viewCount")) else None,
            upload_date=raw_result.get("publishedTime"),
            chnl_name=raw_result["channel"].get("name"),
            chnl_id=raw_result["channel"].get("id"),
        ).__dict__

    @staticmethod
    def format_line(key: str, value: Union[str, int, None]) -> str:
        """Pretty format line"""
        return f"<b>❯  {key}</b> : {value or 'N/A'}"

    @property
    def msg(self) -> str:
        """Message Text"""
        out = f"<a href={YT_VID_URL}{self.yt_id}><b>{escape(self.title)}</b></a>\n"
        if self.body:
            out += f"<pre>{escape(self.body)}</pre>"
        out += "\n".join(
            map(
                lambda x: self.format_line(
                    x.replace("_", " ").title(), getattr(self, x, None)
                ),
                ("duration", "views", "upload_date"),
            )
        )
        out += "\n" + self.format_line(
            "Uploader",
            f"<a href={YT_CHANNEL_URL}{self.chnl_id}>{escape(self.chnl_name)}</a>",
        )
        return out

    def list_view(self, num: int) -> str:
        """List View"""
        return (
            f"<img src={self.thumb}><b><a href={YT_VID_URL}{self.yt_id}>"
            f"{num}. {escape(self.title or 'N/A')}</a></b>"
        )

    def __repr__(self) -> str:
        return json.dumps(self.__dict__, indent=4)


def gen_search_markup(
    key: str, yt_id: str, total: int, page: int = 1
) -> InlineKeyboardMarkup:
    """Get Buttons for search results

    Parameters:
    ----------
        - key (`str`): Search result unique key.
        - yt_id (`str`): YouTube video ID.
        - total (`int`): Total no. of search results.
        - page (`int`, optional): Page number. (Defaults to `1`)

    Returns:
    -------
        `InlineKeyboardMarkup`: for reply_markup
    """

    buttons = [
        [
            InlineKeyboardButton(
                text="⬅️  Back",
                callback_data=f"yt_back|{key}|{page}",
            ),
            InlineKeyboardButton(
                text=f"{page} / {total}",
                callback_data=f"yt_next|{key}|{page}",
            ),
        ],
        [
            InlineKeyboardButton(
                text="📜  List All",
                callback_data=f"yt_listall|{key}",
            ),
            InlineKeyboardButton(
                text="⬇️  Download",
                callback_data=f"yt_extract_info|{yt_id}",
            ),
        ],
    ]
    if page == 1:
        buttons[0].pop(0)
    return InlineKeyboardMarkup(buttons)



__all__ = [
    "run_sync",
    "sublists",
    "humanbytes",
    "time_formater",
    "take_screen_shot",
    "rnd_key",
]

import asyncio
import logging
import os
import tempfile

from datetime import timedelta
from functools import partial, wraps
from io import BytesIO
from pathlib import Path
from random import sample
from typing import Any, Awaitable, Callable, List, Optional, Union

from aiohttp import ClientSession, FormData
from PIL import Image

from iytdl.constants import *  # noqa ignore=F405


_CHAR: List[str] = list("_ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxy")
logger = logging.getLogger(__name__)


def run_sync(func: Callable[..., Any]) -> Awaitable[Any]:
    """Runs the given sync function (optionally with arguments) on a separate thread."""

    @wraps(func)
    async def wrapper(*args: Any, **kwargs: Any):
        return await asyncio.get_running_loop().run_in_executor(
            None, partial(func, *args, **kwargs)
        )

    return wrapper


def sublists(input_list: List[Any], width: int = 3) -> List[List[Any]]:
    """retuns a single list of multiple sublist of fixed width"""
    return [input_list[x : x + width] for x in range(0, len(input_list), width)]


def humanbytes(size: Union[float, int]) -> str:
    """humanize size"""
    if not size:
        return ""
    power = 1024
    t_n = 0
    power_dict = {0: " ", 1: "Ki", 2: "Mi", 3: "Gi", 4: "Ti"}
    while size > power:
        size /= power
        t_n += 1
    return "{:.2f} {}B".format(size, power_dict[t_n])


def time_formater(value: Union[timedelta, int], precision: int = 0) -> str:
    pieces = []
    if isinstance(value, int):
        value = timedelta(seconds=value)
    if value.days:
        pieces.append(f"{value.days}d")

    seconds = value.seconds

    if seconds >= 3600:
        hours = int(seconds / 3600)
        pieces.append(f"{hours}h")
        seconds -= hours * 3600

    if seconds >= 60:
        minutes = int(seconds / 60)
        pieces.append(f"{minutes}m")
        seconds -= minutes * 60

    if seconds > 0 or not pieces:
        pieces.append(f"{seconds}s")

    if precision == 0:
        return " ".join(pieces)

    return " ".join(pieces[:precision])


def rnd_key(length: int = 8) -> str:
    return "".join(sample(_CHAR, length))


async def upload_to_telegraph(http: ClientSession, url: str) -> Optional[str]:
    async with http.get(url) as img_url:
        img_bytes = await img_url.read()
    with tempfile.NamedTemporaryFile(delete=False, suffix=".jpg") as temp_img:
        f_name = temp_img.name
    with Image.open(BytesIO(img_bytes)) as img:
        img.convert("RGB").save(f_name, format="JPEG")
    data = FormData()
    data.add_field(
        name="file",
        value=open(f_name, "rb"),
        content_type="image/jpeg",
        filename="blob",
    )
    async with http.post(
        f"{TELEGRA_PH}/upload",
        headers=TELEGRA_PH_HEADERS,
        data=data,
    ) as resp:
        try:
            return f"{TELEGRA_PH}{(await resp.json())[0].get('src')}"
        except Exception:
            pass
    await run_sync(os.remove)(f_name)


async def take_screen_shot(video_file: str, ttl: int) -> Optional[str]:
    """Generate Thumbnail from video"""
    file = Path(video_file)
    ss_path = file.parent.joinpath(f"{file.stem}.jpg")
    cmd = ["ffmpeg", "-ss", str(ttl), "-i", video_file, "-vframes", "1", str(ss_path)]
    process = await asyncio.create_subprocess_exec(
        *cmd, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
    )
    if stderr := await process.communicate()[1]:
        logger.error(stderr.decode("utf-8", "replace").strip())
    return str(ss_path) if ss_path.is_file() else None



__all__ = ["Uploader"]

import asyncio
import logging
import os

from shutil import rmtree
from typing import Any, Dict, Literal, Optional, Union

from hachoir.metadata import extractMetadata
from hachoir.parser import createParser
from pyrogram import Client
from pyrogram.types import (
    CallbackQuery,
    InputMediaAudio,
    InputMediaDocument,
    InputMediaVideo,
    Message,
)

from iytdl.processes import Process
from iytdl.upload_lib import ext
from iytdl.upload_lib.functions import covert_to_jpg, thumb_from_audio, unquote_filename
from iytdl.upload_lib.progress import progress as upload_progress
from iytdl.utils import *  # noqa ignore=F405


logger = logging.getLogger(__name__)


class Uploader:
    @run_sync
    def find_media(
        self, key: str, media_type: Literal["audio", "video"]
    ) -> Dict[str, Any]:
        """Search Downloaded files for thumbnail and media

        Parameters:
        ----------
            - key (`str`): Unique Key i.e Subfolder name.
            - media_type (`Literal['audio', 'video']`).

        Raises:
        ------
            `TypeError`: In case of Invalid 'media_type'
            `FileNotFoundError`: If Subfolder doesn't exists

        Returns:
        -------
            `Dict[str, Any]`
        """
        if media_type not in ("video", "audio"):
            raise TypeError("'media_type' only accepts video or audio")
        media_path = self.download_path.joinpath(key)
        if not media_path.is_dir():
            raise FileNotFoundError(f"'{media_path}' doesn't exist !")
        info_dict: Dict = {}
        for file in media_path.iterdir():
            if (
                not info_dict.get(media_type)
                and file.name.lower().endswith(getattr(ext, media_type))
                and file.stat().st_size != 0
            ):
                if (
                    file.stat().st_size > 2147000000
                ):  # 2 * 1024 * 1024 * 1024 = 2147483648
                    raise ValueError(
                        f"[{file}] will not be uploaded as filesize exceeds '2 GB' !"
                    )
                f_path = unquote_filename(file)
                info_dict[media_type] = f_path
                info_dict["file_name"] = os.path.basename(f_path)
            if not info_dict.get("thumb") and file.name.lower().endswith(ext.photo):
                info_dict["thumb"], info_dict["size"] = covert_to_jpg(file)

            if media_type in info_dict and "thumb" in info_dict:
                break

        if media := info_dict.get(media_type):
            metadata = extractMetadata(createParser(media))
            if metadata and metadata.has("duration"):
                info_dict["duration"] = metadata.get("duration").seconds

            if media_type == "audio":
                info_dict.pop("size", None)
                if metadata.has("artist"):
                    info_dict["performer"] = metadata.get("artist")
                if metadata.has("title"):
                    info_dict["title"] = metadata.get("title")
                # If Thumb doesn't exist then check for Album art
                if not info_dict.get("thumb"):
                    info_dict["thumb"] = thumb_from_audio(media)
            else:
                width, height = info_dict.pop("size", (1280, 720))
                info_dict["height"] = height
                info_dict["width"] = width
            return info_dict

    async def get_input_media(
        self,
        key: str,
        media_type: Literal["audio", "video"],
        caption: str,
        parse_mode: Optional[str] = "HTML",
    ) -> Union[InputMediaAudio, InputMediaVideo, None]:
        """Get Input Media

        Parameters:
        ----------
            - key (`str`): Unique Key.
            - media_type (`Literal['audio', 'video']`): audio or video.
            - caption (`str`): Media caption text.
            - parse_mode (`Optional[str]`, optional):
                By default, texts are parsed using both Markdown and HTML styles.
                You can combine both syntaxes together.
                Pass "markdown" or "md" to enable Markdown-style parsing only.
                Pass "html" to enable HTML-style parsing only.
                Pass None to completely disable style parsing. (Defaults to `"HTML"`)

        Returns:
        -------
            `Union[InputMediaAudio, InputMediaVideo, None]`
        """
        if media_kwargs := await self.find_media(key, media_type):
            media_kwargs.update(
                {
                    "media": media_kwargs.pop(media_type),
                    "caption": caption,
                    "parse_mode": parse_mode,
                }
            )
            if media_type == "audio":
                return InputMediaAudio(**media_kwargs)
            if media_type == "video":
                return InputMediaVideo(**media_kwargs)

    async def upload(
        self,
        client: Client,
        key: str,
        downtype: str,
        update: Union[CallbackQuery, Message],
        caption_link: Optional[str] = None,
        with_progress: bool = True,
    ) -> Union[CallbackQuery, Message]:
        """Upload downloaded Media with progress

        Parameters:
        ----------
            - client (`Client`): Pyrogram Bot Client.
            - key (`str`): Unique key to find downloaded media.
            - downtype (`str`): (`Literal['audio', 'video']`).
            - update (`Union[CallbackQuery, Message]`): Pyrogram Update to edit message.
            - caption_link (`Optional[str]`, optional): Custom caption href link. (Defaults to `None`)
            - with_progress (`bool`, optional): Enable / Disable progress. (Defaults to `True`)

        Returns:
        -------
            `Union[CallbackQuery, Message]`: On Success
        """
        if mkwargs := await self.find_media(key, downtype):

            if caption_link:
                caption = f"<b><a href={caption_link}>{mkwargs['file_name']}</a></b>"
            else:
                caption = f"<b>{mkwargs['file_name']}</b>"
            process = Process(update)
            try:
                if downtype == "video":
                    return await self.__upload_video(
                        client, process, caption, mkwargs, with_progress
                    )
                if downtype == "audio":
                    return await self.__upload_audio(
                        client, process, caption, mkwargs, with_progress
                    )
            finally:
                if self.delete_file_after_upload:
                    rmtree(self.download_path.joinpath(key), ignore_errors=True)

    async def __upload_video(
        self,
        client: Client,
        process: Process,
        caption: str,
        mkwargs: Dict[str, Any],
        with_progress: bool = True,
    ):

        if not mkwargs.get("thumb") and (duration := mkwargs.get("duration")):
            ttl = duration // 2

            mkwargs["thumb"] = await take_screen_shot(mkwargs["video"], ttl)

        if not (
            uploaded := await client.send_video(
                chat_id=self.log_group_id,
                caption=f"📹  {caption}",
                parse_mode="HTML",
                disable_notification=True,
                progress=upload_progress if with_progress else None,
                progress_args=(client, process, mkwargs["file_name"])
                if with_progress
                else (),
                **mkwargs,
            )
        ):

            return
        await asyncio.sleep(2)
        if not process.is_cancelled:
            if uploaded.video:

                return await process.edit_media(
                    media=InputMediaVideo(
                        uploaded.video.file_id, caption=uploaded.caption.html
                    ),
                    reply_markup=None,
                )
            elif uploaded.document:
                return await process.edit_media(
                    media=InputMediaDocument(
                        uploaded.document.file_id, caption=uploaded.caption.html
                    ),
                    reply_markup=None,
                )

    async def __upload_audio(
        self,
        client: Client,
        process: Process,
        caption: str,
        mkwargs: Dict[str, Any],
        with_progress: bool = True,
    ):
        if not (
            uploaded := await client.send_audio(
                chat_id=self.log_group_id,
                caption=f"🎵  {caption}",
                parse_mode="HTML",
                disable_notification=True,
                progress=upload_progress if with_progress else None,
                progress_args=(client, process, mkwargs["file_name"])
                if with_progress
                else (),
                **mkwargs,
            )
        ):
            return
        await asyncio.sleep(2)
        if not process.is_cancelled:
            if uploaded.audio:
                return await process.edit_media(
                    media=InputMediaAudio(
                        uploaded.audio.file_id, caption=uploaded.caption.html
                    ),
                    reply_markup=None,
                )
            elif uploaded.document:
                return await process.edit_media(
                    media=InputMediaDocument(
                        uploaded.document.file_id, caption=uploaded.caption.html
                    ),
                    reply_markup=None,
                )



__all__ = ["SearhResult"]

import json

from pyrogram.types import InlineKeyboardMarkup


class SearhResult:
    key: str
    buttons: InlineKeyboardMarkup
    text: str
    image: str

    def __init__(
        self,
        key: str,
        text: str,
        image: str,
        buttons: InlineKeyboardMarkup,
    ) -> None:
        self.key = key
        self.buttons = buttons
        self.caption = text
        self.image_url = image

    def __repr__(self) -> str:
        out = self.__dict__.copy()
        out["buttons"] = (
            json.loads(str(btn)) if (btn := out.pop("buttons", None)) else None
        )
        return json.dumps(out, indent=4)



__all__ = ["progress"]

import asyncio
import logging
import time

from math import floor
from typing import Dict, Tuple

from pyrogram import Client, ContinuePropagation, StopPropagation, StopTransmission
from pyrogram.errors import FloodWait, MessageNotModified

from iytdl.processes import Process
from iytdl.utils import *


_PROGRESS: Dict[str, Tuple[int, int]] = {}
LOG = logging.getLogger(__name__)


async def progress(
    current: int,
    total: int,
    client: Client,
    process: Process,
    filename: str,
    mode: str = "upload",
    edit_rate: int = 8,
):
    """Pyrogram Upload / Download Progress Bar

    Parameters:
    ----------
        - current (`int`): The amount of bytes transmitted so far.
        - total (`int`): The total size of the file.
        - client (`Client`): Pyrogram Client.
        - process (`Process`): iytdl.process.Process.
        - filename (`str`): Display name of the file.
        - mode (`str`, optional): `"upload"` or `"download"`. (Defaults to `"upload"`)
        - edit_rate (`int`, optional): Message edit rate. (Defaults to `8`)
    """
    if process.is_cancelled:
        LOG.warning("Upload process is Cancelled")
        # Stop Uploading
        await client.stop_transmission()

    if current == total:
        if process.id not in _PROGRESS:
            return
        del _PROGRESS[process.id]
        try:
            await process.edit(f"`Finalizing {mode} process ...`")
        except FloodWait as f_w:
            await asyncio.sleep(f_w.x + 1)
        return
    now = int(time.time())
    if process.id not in _PROGRESS:
        _PROGRESS[process.id] = (now, now)
    start, last_update_time = _PROGRESS[process.id]
    # ------------------------------------ #
    if (now - last_update_time) >= edit_rate:
        _PROGRESS[process.id] = (start, now)
        # Only edit message once every 8 seconds to avoid ratelimits
        after = now - start
        speed = current / after
        eta = round((total - current) / speed)
        percentage = round(current / total * 100)
        progress_bar = (
            f"[{'█' * floor(15 * percentage / 100)}"
            f"{'░' * floor(15 * (1 - percentage / 100))}]"
        )
        progress = f"""
<i>{mode.title()}ing:</i>  <code>{filename}</code>
<b>Completed:</b>  <code>{humanbytes(current)} / {humanbytes(total)}</code>
<b>Progress:</b>  <code>{progress_bar} {percentage} %</code>
<b>Speed:</b>  <code>{humanbytes(speed)}</code>
<b>ETA:</b>  <code>{time_formater(eta)}</code>
"""
        try:
            await process.edit(progress, reply_markup=process.cancel_markup)
        except FloodWait as f:
            await asyncio.sleep(f.x + 2)
        except (ContinuePropagation, MessageNotModified):
            pass
        except (StopPropagation, StopTransmission) as p_e:
            raise p_e
        except Exception:
            LOG.exception("Unable to Edit message")



__all__ = ["unquote_filename", "thumb_from_audio", "covert_to_jpg"]

import re

from io import BytesIO
from pathlib import Path
from typing import Optional, Tuple, Union

import mutagen

from PIL import Image

from iytdl.upload_lib import ext


def unquote_filename(filename: Union[Path, str]) -> str:
    """
    Removes single and double quotes from filename to avoid ffmpeg errors
    due to unclosed quotation in filename

    Parameters:
    ----------
        filename (`Union[Path, str]`): Full file name.

    Returns:
    -------
        str: New filename after renaming original file

    """
    file = Path(filename) if isinstance(filename, str) else filename
    un_quoted = file.parent.joinpath(re.sub(r"[\"']", "", file.name))
    if file.name != un_quoted.name:
        file.rename(un_quoted)
        return str(un_quoted)
    return str(filename)


def thumb_from_audio(filename: Union[Path, str]) -> Optional[str]:
    """Extract album art from audio

    Parameters:
    ----------
        filename (`Union[Path, str]`): audio file path.

    Returns:
    -------
        Optional[str]: if audio has album art

    """
    file = Path(filename) if isinstance(filename, str) else filename
    if not (audio_id3 := mutagen.File(str(file))):
        return
    thumb_path = file.parent.joinpath("album_art.jpg")
    for key in audio_id3.keys():
        if "APIC" in key and (album_art := getattr(audio_id3[key], "data", None)):
            thumb_path = file.parent.joinpath("album_art.jpg")
            with BytesIO(album_art) as img_io:
                with Image.open(img_io) as img:
                    img.convert("RGB").save(str(thumb_path), "JPEG")
            break
    if thumb_path.is_file():
        return str(thumb_path)


def covert_to_jpg(filename: Union[Path, str]) -> Tuple[str, Tuple[int]]:
    """Convert images to Telegram supported thumb

    Parameters:
    ----------
        filename (`Union[Path, str]`): Image file path.

    Returns:
    -------
        `Tuple[str, Tuple[int]]`: (thumb_path, dimensions)

    """
    file = Path(filename) if isinstance(filename, str) else filename
    with Image.open(file) as img:
        if file.name.lower().endswith(ext.photo[:2]):
            thumb_path = str(file)
        else:
            thumb_path = str(file.parent.joinpath(f"{file.stem}.jpeg"))
            img.convert("RGB").save(thumb_path, "JPEG")
        size = img.size
    return thumb_path, size



__all__ = ["ExternalDownloader", "Aria2c"]

from dataclasses import dataclass
from typing import Dict, List, Union


class ExternalDownloader:
    def _export(self) -> Dict[str, Union[str, List[str]]]:
        attrs = list(self.__dataclass_fields__)
        if "executable_path" in attrs:
            ext_dl = self.executable_path
            attrs.remove("executable_path")
        else:
            ext_dl = self.__class__.__name__.lower()

        return dict(
            external_downloader=ext_dl,
            external_downloader_args=list(
                map(
                    lambda x: f"--{x.replace('_', '-')}={getattr(self, x)}",
                    attrs,
                )
            ),
        )


@dataclass
class Aria2c(ExternalDownloader):
    """Aria2c External Downloader"""

    executable_path: str = "aria2c"
    max_concurrent_downloads: int = 5
    max_connection_per_server: int = 1
    split: int = 5
    min_split_size: str = "20M"



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['iytdl', 'iytdl.types', 'iytdl.upload_lib']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=8.3.1,<9.0.0',
 'Pyrogram>=1.2.9,<2.0.0',
 'TgCrypto>=1.2.2,<2.0.0',
 'aiohttp>=3.7.4,<4.0.0',
 'aiosqlite>=0.17.0,<0.18.0',
 'hachoir>=3.1.2,<4.0.0',
 'html-telegraph-poster>=0.4.0,<0.5.0',
 'mutagen>=1.45.1,<2.0.0',
 'youtube-search-python==1.4.6',
 'yt-dlp>=2021.8.10,<2022.0.0']

setup_kwargs = {
    'name': 'iytdl',
    'version': '0.4.1',
    'description': 'Aysnc Inline YouTube-DL for Pyrogram',
    'long_description': '<p align="center">\n<img src="https://i.imgur.com/Q94CDKC.png" width=250px>\n\n# iYTDL\n\n<a href="https://github.com/iytdl/iytdl/blob/main/LICENSE"><img alt="License: GPLv3" src="https://img.shields.io/badge/License-GPLv3-blue.svg"></a>\n<a href="https://github.com/iytdl/iytdl/actions"><img alt="Actions Status" src="https://github.com/iytdl/iytdl/actions/workflows/pypi-publish.yaml/badge.svg"></a>\n<a href="https://pypi.org/project/iytdl/"><img alt="PyPI" src="https://img.shields.io/pypi/v/iytdl"></a>\n<a href="https://pepy.tech/project/iytdl"><img alt="Downloads" src="https://pepy.tech/badge/iytdl"></a>\n<a href="https://github.com/psf/black"><img alt="Code style: black" src="https://img.shields.io/badge/code%20style-black-000000.svg"></a>\n\n</p>\n\n<h2 align="center"> Aysnc Inline YouTube-DL for Pyrogram</h2>\n\n## ⬇️ Installation\n\n> Install\n\n```bash\npip3 install iytdl\n```\n\n> Upgrade\n\n```bash\npip3 install -U iytdl\n```\n\n> Build Wheel Locally\n\n```bash\ngit clone https://github.com/iytdl/iytdl.git\ncd iytdl\npoetry install\n\nchmod +x scripts/install.sh && ./scripts/install.sh\n```\n\n## Features\n\n- Async and memory efficient (uses Aiosqlite for Caching)\n- Uses hashing avoid storing duplicate data\n- Supports context manager\n- Supports External Downloader [[Aria2c](https://github.com/iytdl/iytdl/blob/master/tests/test_download_upload.py#L20)]\n- [Supported Sites](https://github.com/yt-dlp/yt-dlp/blob/master/supportedsites.md)\n\n## Requirements\n\n- [YT-DLP](https://github.com/yt-dlp/yt-dlp) (Active youtube-dl fork)\n- [Python](https://www.python.org/) >=3.8,<4\n- [Pyrogram](https://docs.pyrogram.org/) based Bot\n- [FFmpeg](http://ffmpeg.org/)\n- [Aria2c](https://aria2.github.io/) (_Optional_)\n\n## Pre-commit Hooks\n\n- [Install Pre-commit Hooks](https://pre-commit.com/#installation)\n- `pre-commit install`\n\n## Examples\n\n### Callbacks\n\n<details>\n  <summary><b>OPEN</b></summary>\n\n- Back and Next\n\n```python\nr"^yt_(back|next)\\|(?P<key>[\\w-]{5,11})\\|(?P<pg>\\d+)$"\n```\n\n- List View\n\n```python\nr"^yt_listall\\|(?P<key>[\\w-]{5,11})$"\n```\n\n- Extract Info\n\n```python\nr"^yt_extract_info\\|(?P<key>[\\w-]{5,11})$"\n```\n\n- Download\n\n```python\nr"yt_(?P<mode>gen|dl)\\|(?P<key>[\\w-]+)\\|(?P<choice>[\\w-]+)\\|(?P<dl_type>a|v)$"\n```\n\n- Cancel\n\n```python\nr"^yt_cancel\\|(?P<process_id>[\\w\\.]+)$"\n```\n\n</details>\n\n### Module\n\n### [YouTube.py](https://github.com/code-rgb/droid/blob/master/droid/modules/youtube.py)\n',
    'author': 'Leorio Paradinight',
    'author_email': '62891774+code-rgb@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/iytdl/iytdl',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4',
}


setup(**setup_kwargs)



__all__ = ["Process"]

from typing import Callable, Set, Union

from pyrogram.types import (
    CallbackQuery,
    InlineKeyboardButton,
    InlineKeyboardMarkup,
    Message,
)

from iytdl.exceptions import UnsupportedUpdateError


class Process:
    cancelled_ids: Set[str] = set()

    def __init__(self, update: Union[Message, CallbackQuery]) -> None:
        """
        Parameters:
        ----------
            update (`Union[Message, CallbackQuery]`)

        Raises:
        ------
            UnsupportedUpdateError: In case of unsupported update.
        """

        if not isinstance(update, (Message, CallbackQuery)):
            raise UnsupportedUpdateError

        if msg := (update if isinstance(update, Message) else update.message):
            process_id = f"{msg.chat.id}.{msg.message_id}"
            edit_func = msg.edit_text
            media_edit_func = msg.edit_media
        else:
            process_id = str(update.id)
            edit_func = update.edit_message_text
            media_edit_func = update.edit_message_media

        self.edit: Callable = edit_func
        self.edit_media: Callable = media_edit_func
        self.id: str = process_id

    @classmethod
    def cancel_id(cls, process_id: str) -> None:
        """Cancel Upload / Download Process by ID

        Parameters:
        ----------
            process_id (`str`): Unique ID.

        """
        cls.cancelled_ids.add(process_id)

    @classmethod
    def remove_id(cls, process_id: str) -> None:
        """Remove cancelled ID

        Parameters:
        ----------
            process_id (`str`): Unique ID.

        """
        cls.cancelled_ids.remove(process_id)

    @property
    def cancel(self) -> None:
        """Cancel process"""
        self.cancelled_ids.add(self.id)

    @property
    def is_cancelled(self) -> bool:
        """Check if process is cancelled

        Returns:
        -------
            bool: True if cancelled else False
        """
        return self.id in self.cancelled_ids

    @property
    def cancel_markup(self) -> InlineKeyboardMarkup:
        return InlineKeyboardMarkup(
            [[InlineKeyboardButton("❌ Cancel", callback_data=f"yt_cancel|{self.id}")]]
        )



__all__ = ["Extractor"]

import logging

from collections import defaultdict
from typing import Dict, List, Optional, Tuple

import yt_dlp as youtube_dl

from pyrogram.types import InlineKeyboardButton, InlineKeyboardMarkup

# from youtube_dl.utils import DownloadError, ExtractorError, UnsupportedError
from yt_dlp.utils import DownloadError, ExtractorError, UnsupportedError

from iytdl.constants import YT_VID_URL
from iytdl.formatter import ResultFormatter as res_f
from iytdl.types import SearhResult
from iytdl.utils import *  # noqa ignore=F405


logger = logging.getLogger(__name__)


class Extractor:
    def __init__(self, silent: bool = False) -> None:
        self.silent = silent

    @run_sync
    def generic_extractor(self, key: str, url: str) -> Optional[SearhResult]:
        """Generic extractor for URLs other than YouTube
        [more info](https://github.com/yt-dlp/yt-dlp/blob/master/supportedsites.md).

        Parameters:
        ----------
            - key (`str`): Unique key for Callback.
            - url (`str`): Http URL.

        Returns:
        -------
            `Optional[SearhResult]`: On Success
        """
        # passing key as we can't pass the entire url in callback_data
        buttons = [
            [
                InlineKeyboardButton(
                    "⭐️ BEST - 📹 Video", callback_data=f"yt_gen|{key}|mp4|v"
                ),
                InlineKeyboardButton(
                    "⭐️ BEST - 🎧 Audio", callback_data=f"yt_gen|{key}|mp3|a"
                ),
            ]
        ]
        params = {"no-playlist": True, "quiet": self.silent, "logtostderr": self.silent}
        try:
            resp = youtube_dl.YoutubeDL(params).extract_info(url, download=False)
            # with open("j_debug_data.json", "w") as fx:
            #     json.dump(resp, fx, indent=4, sort_keys=True)
        except UnsupportedError:
            logger.error(f"[URL -> {url}] - is not NOT SUPPORTED")
            return
        except DownloadError as d_e:
            logger.error(f"[URL -> {url}] - {d_e}")
            return
        except ExtractorError:
            logger.warning(f"[URL -> {url}] - Failed to Extract Info")
            return SearhResult(
                key,
                "[No Information]",
                self.default_thumb,
                InlineKeyboardMarkup(buttons),
            )
        msg = f"<b><a href={url}>{resp.get('title', '[No Title]')}</a></b>\n"
        if description := resp.get("description"):
            msg += (
                f"<pre>{description[:380]}...</pre>\n"
                if len(description) > 380
                else f"<pre>{description}</pre>\n"
            )
        for info_type in ("duration", "uploader"):
            if info := resp.get(info_type):
                msg += f"{res_f.format_line(info_type.title(), info)}\n"
        if formats := (
            resp.get("formats")
            or (
                (
                    entries[0].get("formats")
                    if (entries := resp.get("entries")) and len(entries) != 0
                    else None
                )
                if resp.get("_type", "") == "playlist"
                else None
            )
        ):
            buttons += sublists(
                list(
                    map(
                        lambda x: InlineKeyboardButton(
                            " | ".join(
                                filter(
                                    None,
                                    (
                                        x.get("format"),
                                        x.get("ext"),
                                        humanbytes(v_file_size)
                                        if (v_file_size := x.get("filesize"))
                                        else None,
                                    ),
                                )
                            ),
                            callback_data=f"yt_gen|{key}|{x.get('format_id')}|v",
                        ),
                        self.filter_generic_formats(formats),
                    )
                ),
                width=1,
            )
        return SearhResult(
            key,
            msg[:1020],
            resp.get("thumbnail", self.default_thumb),
            InlineKeyboardMarkup(buttons),
        )

    def filter_generic_formats(self, raw_formats: Dict) -> Dict:
        """Filter Formats Based on 'tbr', 'width' and 'acodec'

        Parameters:
        ----------
            - raw_formats (`Dict`): Raw youtube-dl formats.

        Returns:
        -------
            `Dict`: Filtered Formats
        """
        widthset = set()

        def qual_filter(frmt) -> bool:
            if frmt.get("tbr") and frmt.get("acodec") and (width := frmt.get("width")):
                if not (width in widthset):
                    widthset.add(width)
                    return True
            return False

        frmt_list = list(
            filter(
                qual_filter,
                sorted(
                    raw_formats, key=lambda x: float(x.get("tbr") or 0), reverse=True
                ),
            )
        )[:25]
        return frmt_list if len(frmt_list) > 1 else raw_formats

    @run_sync
    def get_download_button(self, yt_id: str) -> SearhResult:
        """Generate Inline Buttons for YouTube Video

        Parameters:
        ----------
            - yt_id (`str`): YouTube video key.

        Returns:
        -------
            `SearhResult`: ~iytdl.types.SearhResult
        """
        buttons = [
            [
                InlineKeyboardButton(
                    "⭐️ BEST - 📹 MKV", callback_data=f"yt_dl|{yt_id}|mkv|v"
                ),
                InlineKeyboardButton(
                    "⭐️ BEST - 📹 WebM/MP4",
                    callback_data=f"yt_dl|{yt_id}|mp4|v",
                ),
            ]
        ]
        best_audio_btn = [
            [
                InlineKeyboardButton(
                    "⭐️ BEST - 🎵 320Kbps - MP3",
                    callback_data=f"yt_dl|{yt_id}|mp3|a",
                )
            ]
        ]
        params = {"no-playlist": True, "quiet": self.silent, "logtostderr": self.silent}
        try:
            vid_data = youtube_dl.YoutubeDL(params).extract_info(
                f"{YT_VID_URL}{yt_id}", download=False
            )
        except ExtractorError:
            vid_data = None
            buttons += best_audio_btn
        else:
            # ------------------------------------------------ #
            qual_dict = defaultdict(lambda: defaultdict(int))
            qual_list = ("1440p", "1080p", "720p", "480p", "360p", "240p", "144p")
            audio_dict: Dict[int, str] = {}
            # ------------------------------------------------ #
            for video in vid_data["formats"]:
                fr_note = video.get("format_note")
                fr_id = int(video.get("format_id"))
                fr_size = video.get("filesize")
                if video.get("ext") == "mp4":
                    for frmt_ in qual_list:
                        if fr_note in (frmt_, frmt_ + "60"):
                            qual_dict[frmt_][fr_id] = fr_size
                if video.get("acodec") != "none":
                    bitrrate = int(video.get("abr", 0))
                    if bitrrate != 0:
                        audio_dict[
                            bitrrate
                        ] = f"🎵 {bitrrate}Kbps ({humanbytes(fr_size) or 'N/A'})"
            video_btns: List[InlineKeyboardButton] = []
            for frmt in qual_list:
                frmt_dict = qual_dict[frmt]
                if len(frmt_dict) != 0:
                    frmt_id = sorted(list(frmt_dict))[-1]
                    frmt_size = humanbytes(frmt_dict.get(frmt_id)) or "N/A"
                    video_btns.append(
                        InlineKeyboardButton(
                            f"📹 {frmt} ({frmt_size})",
                            callback_data=f"yt_dl|{yt_id}|{frmt_id}|v",
                        )
                    )
            buttons += sublists(video_btns, width=2)
            buttons += best_audio_btn
            buttons += sublists(
                list(
                    map(
                        lambda x: InlineKeyboardButton(
                            audio_dict[x], callback_data=f"yt_dl|{yt_id}|{x}|a"
                        ),
                        sorted(audio_dict.keys(), reverse=True),
                    )
                ),
                width=2,
            )

        return SearhResult(
            yt_id,
            (
                f"<b><a href={YT_VID_URL}{yt_id}>{vid_data.get('title')}</a></b>"
                if vid_data
                else ""
            ),
            vid_data.get("thumbnail") if vid_data else self.default_thumb,
            InlineKeyboardMarkup(buttons),
        )

    @staticmethod
    def get_choice_by_id(
        choice_id: str, media_type: str, yt_url: bool = True
    ) -> Tuple[str]:
        """youtube-dl downloader formats for video / audio

        Parameters:
        ----------
            - choice_id (`str`): Format choice.
            - media_type (`str`): `"video"` or `"audio"`.
            - yt_url (`bool`, optional): If URL is from https://www.youtube.com/. (Defaults to `True`)

        Returns:
        -------
            `Tuple[str]`: (Full Format str, Display str)
        """
        if choice_id == "mkv":
            # Download and Merge (best video) + (best audio)
            choice_str = "(bestvideo+bestaudio/best)[filesize<?1950M]"
            disp_str = "Best (video+audio)"
        elif choice_id == "mp4":
            # Download best Mp4 format if available
            # Webm is not supported by Telegram anymore :(
            if yt_url:
                choice_str = "(bestvideo[ext=mp4]+(258/256/bestaudio[ext=m4a])/best[ext=mp4]/best)[filesize<?1950M]"
                disp_str = "Best MP4"
            else:
                choice_str = "(bestvideo[ext=mp4]+bestaudio[ext=m4a]/best[ext=mp4]/best)[filesize<?1950M]"
                disp_str = "Best MP4 (video+audio)"
        elif choice_id == "mp3":
            choice_str = "320"
            disp_str = "320 Kbps"
        else:
            disp_str = choice_id
            if media_type == "v":
                # mp4 video quality + best compatible audio
                if yt_url:
                    choice_str = f"({disp_str}+(258/256/bestaudio[ext=m4a])/best[ext=mp4]/best)[filesize<?1950M]"
                else:
                    choice_str = f"({disp_str}+bestaudio[ext=m4a]/best[ext=mp4]/best)[filesize<?1950M]"
            else:  # Audio
                choice_str = disp_str
        return choice_str, disp_str
